"""package in order to reduce and analyze fits

:author: Clement Hottier
"""

from .utils import (
    alignstack,
    clean_cosmic,
    crosscorrelalign,
    makemasterbias,
    makemasterflat,
    processimages,
)
from .__about__ import __version__

__author__ = "Clement Hottier, Noel Robichon"
__all__ = [
    "makemasterbias",
    "makemasterflat",
    "processimages",
    "crosscorrelalign",
    "alignstack",
    "clean_cosmic",
]
